
from ..runenvtool import RunEnvTool


class bzip2Tool(RunEnvTool):
    """Freely available, patent free (see below), high-quality data compressor.

Home: http://www.bzip.org/
"""
    __slots__ = ()

    def _installTool(self, env):
        self._install.debrpm(['bzip2'])
        self._install.emerge(['app-arch/bzip2'])
        self._install.pacman(['bzip2'])
        self._install.apk(['bzip2'])
        self._install.brew('bzip2')

    def initEnv(self, env, bin_name=None):
        # Busybox's version is not enough for SDKMan
        if self._detect.isAlpineLinux() and self._ospath.islink('/usr/bin/bzip2'):
            return

        super(bzip2Tool, self).initEnv(env, bin_name)




from ..runenvtool import RunEnvTool


class curlTool(RunEnvTool):
    """Command line tool and library for transferring data with URLs.

Home: https://curl.haxx.se/
"""
    __slots__ = ()

    def _installTool(self, env):
        self._install.debrpm(['curl'])
        self._install.emerge(['net-misc/curl'])
        self._install.pacman(['curl'])
        self._install.apk(['curl'])
        self._install.brew('curl')




class CurlToolMixIn(object):
    __slots__ = ()

    def getDeps(self):
        return ['curl']

    def _callCurl(self, env, curl_args, *args, **nargs):
        timeout_args = self._configutil.timeouts(env, 'curl')

        return self._executil.callExternal(
            [env['curlBin'], '-fsSL'] + timeout_args + curl_args,
            *args, **nargs
        )




from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn


class webpackTool(NpmToolMixIn, BuildTool):
    """webpack is a module bundler (JS world)

Home: https://webpack.js.org
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return 'webpack.config.js'

    def onBuild(self, config):
        webpackBin = config['env']['webpackBin']
        self._executil.callExternal([
            webpackBin, '-p', '--profile', '--display', '--verbose'])




from ..util.log import *

__all__ = ('LogMixIn')


class LogMixIn(object):
    __slots__ = ()

    def _info(self, msg, label=None):
        if label:  # for backward compatibility
            infoLabel(label, msg)
        else:
            info(msg)

    def _infoLabel(self, label, msg):
        infoLabel(label, msg)

    def _warn(self, msg):
        warn(msg)

    def _errorExit(self, msg):
        errorExit(msg)




from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn


class yarnTool(NpmToolMixIn, BuildTool):
    """YARN: fast, reliable, and secure dependency management.

Home: https://yarnpkg.com

Note: auto-detected only if yarn.lock is present
"""
    __slots__ = ()

    def getOrder(self):
        # required to run before other npm-based tools
        return -10

    def autoDetectFiles(self):
        return 'yarn.lock'

    def onPrepare(self, config):
        yarnBin = config['env']['yarnBin']
        self._executil.callExternal([yarnBin, 'install'])

    def onPackage(self, config):
        yarnBin = config['env']['yarnBin']
        self._executil.callExternal([yarnBin, 'install', '--production'])




from ..buildtool import BuildTool
from ..testtool import TestTool
from .sdkmantoolmixin import SdkmanToolMixIn


class mavenTool(SdkmanToolMixIn, BuildTool, TestTool):
    """Apache Maven is a software project management and comprehension tool.

Home: https://maven.apache.org/

Expects clean, compile, package and test targets.

Build targets:
    prepare -> clean
    build -> compile
    package -> package
    check -> test
Override targets with .config.toolTune.

Requires Java >= 7.
"""
    __slots__ = ()

    def _minJava(self):
        return '7'

    def autoDetectFiles(self):
        return 'pom.xml'

    def _binName(self):
        return 'mvn'

    def onPrepare(self, config):
        target = self._getTune(config, 'prepare', 'clean')
        self._executil.callExternal([config['env']['mavenBin'], target])

    def onBuild(self, config):
        target = self._getTune(config, 'build', 'compile')
        self._executil.callExternal([config['env']['mavenBin'], target])

    def onPackage(self, config):
        target = self._getTune(config, 'package', 'package')
        self._executil.callExternal([config['env']['mavenBin'], target])
        self._pathutil.addPackageFiles(config, 'target/*.jar')

    def onCheck(self, config):
        target = self._getTune(config, 'check', 'test')
        self._executil.callExternal([config['env']['mavenBin'], target])




from ..runenvtool import RunEnvTool


class gpgTool(RunEnvTool):
    """The GNU Privacy Guard.

Home: https://www.gnupg.org/

gpgKeyServer is hkp://keyserver.ubuntu.com:80 by default
"""
    __slots__ = ()

    def envNames(self):
        return ['gpgBin', 'gpgKeyServer']

    def _installTool(self, env):
        self._install.debrpm(['gnupg', 'gnupg2'])
        self._install.emerge(['app-crypt/gnupg'])
        self._install.pacman(['gnupg'])
        self._install.apk(['gnupg'])
        self._install.brew('gnupg')

    def initEnv(self, env):
        super(gpgTool, self).initEnv(env)
        env.setdefault('gpgKeyServer', 'hkp://keyserver.ubuntu.com:80')




class JavaToolMixIn(object):
    __slots__ = ()

    _LATEST_JAVA = '8'

    def _minJava(self):
        return None

    def envDeps(self, env):
        min_java = self._minJava()

        if not min_java:
            return

        for ek in ['javaVer', 'jdkVer']:
            if env.get(ek, None):
                if int(env[ek]) < int(min_java):
                    self._errorExit('{0} requires minimal {1}={2}'
                                    .format(self._name, ek, min_java))
            else:
                env[ek] = str(self._LATEST_JAVA)




from ..buildtool import BuildTool


class releaseTool(BuildTool):
    """FutoIn CID-specific release processing

Python: replaces "*__version__"' with "__version__ = '{version}'"
Changelog: replaces "=== \(next\) ===" with "=== {version} ({date}) ==="

Tune:
    .python = [] - list of Python files
    .changelog = 'CHANGELOG.txt' - list of ChangeLog files
"""
    __slots__ = ()

    def autoDetect(self, config):
        return True

    def updateProjectConfig(self, config, updates):
        re = self._ext.re

        def py_updater(content):
            if 'version' in updates:
                return re.sub(
                    r'^.*__version__.*$',
                    '__version__ = \'{0}\''.format(updates['version']),
                    content,
                    flags=re.MULTILINE
                )

        def cl_updater(content):
            if 'version' in updates:
                date = self._ext.datetime.datetime.utcnow().isoformat().split('T')[
                    0]
                return re.sub(
                    r'^=== \(next\) ===$',
                    '=== {0} ({1}) ==='.format(updates['version'], date),
                    content,
                    count=1,
                    flags=re.MULTILINE
                )

        res = []
        #---
        py_files = self._getTune(config, 'python', [])
        py_files = self._configutil.listify(py_files)

        for pyf in py_files:
            res += self._pathutil.updateTextFile(pyf, py_updater)

        #---
        cl_files = self._getTune(config, 'changelog', 'CHANGELOG.txt')
        cl_files = self._configutil.listify(cl_files)

        for clf in cl_files:
            res += self._pathutil.updateTextFile(clf, cl_updater)

        #---
        return res

    def initEnv(self, env):
        self._have_tool = True




from ..buildtool import BuildTool


class makeTool(BuildTool):
    """GNU Make.

Home: https://www.gnu.org/software/make/

Expects presence of "clean" target.
Build uses the default target.
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return [
            'GNUmakefile',
            'makefile',
            'Makefile',
        ]

    def _installTool(self, env):
        self._install.debrpm(['make'])
        self._install.emerge(['sys-devel/make'])
        self._install.pacman(['make'])
        self._install.apk(['make'])

    def onPrepare(self, config):
        self._executil.callExternal([config['env']['makeBin'], 'clean'])

    def onBuild(self, config):
        target = self._getTune(config, 'build')

        if target:
            args = [target]
        else:
            args = []

        self._executil.callExternal([config['env']['makeBin']] + args)




from ..runenvtool import RunEnvTool


class futoinTool(RunEnvTool):
    """futoin.json updater as defined in FTN16.

Home: https://github.com/futoin/specs/blob/master/draft/ftn16_cid_tool.md

futoin.json is the only file read by FutoIn CID itself.
"""
    __slots__ = ()
    _FUTOIN_JSON = 'futoin.json'

    def autoDetectFiles(self):
        return self._FUTOIN_JSON

    def initEnv(self, env):
        self._have_tool = True

    def updateProjectConfig(self, config, updates):
        def updater(json):
            for f in ('name', 'version'):
                if f in updates:
                    json[f] = updates[f]

        return self._pathutil.updateJSONConfig(self._FUTOIN_JSON, updater)




from ..buildtool import BuildTool
from .gemtoolmixin import GemToolMixIn


class bundlerTool(GemToolMixIn, BuildTool):
    """Bundler: The best way to manage a Ruby application's gems.

Home: http://bundler.io/
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return 'Gemfile'

    def initEnv(self, env):
        super(bundlerTool, self).initEnv(env, 'bundle')

    def onPrepare(self, config):
        self._executil.callExternal([config['env']['bundlerBin'], 'install'])

    def onPackage(self, config):
        self._executil.callExternal([config['env']['bundlerBin'], 'install',
                                     '--deployment', '--clean'])




from ..runenvtool import RunEnvTool


class zipTool(RunEnvTool):
    """package and compress (archive) files.
"""
    __slots__ = ()

    def _installTool(self, env):
        self._install.debrpm(['zip'])
        self._install.emerge(['app-arch/zip'])
        self._install.pacman(['zip'])
        self._install.apk(['zip'])




from ...mixins.ondemand import ext as _ext


def emerge(packages):
    if not _ext.detect.isEmerge():
        return

    emerge = _ext.pathutil.which('emerge')

    packages = _ext.configutil.listify(packages)

    _ext.executil.trySudoCall(
        [emerge] + packages,
        errmsg='you may need to install the build deps manually !'
    )


def emergeDepsOnly(packages):
    packages = _ext.configutil.listify(packages)
    emerge(['--onlydeps'] + packages)




from ..runenvtool import RunEnvTool


class gccTool(RunEnvTool):
    """GNU Compiler Collection.

Home: https://gcc.gnu.org/

If gccBin is set it must point to bin folder.
"""
    __slots__ = ()

    def envNames(self):
        return ['gccBin', 'gccPrefix', 'gccPostfix']

    def _installTool(self, env):
        self._install.deb(['gcc'])
        self._install.rpm(['gcc'])
        self._install.emerge(['sys-devel/gcc'])
        self._install.pacman(['gcc'])
        self._install.apk(['gcc'])

    def initEnv(self, env):
        gcc_bin = env.get('gccBin', None)

        pref = env.setdefault('gccPrefix', '')
        postf = env.setdefault('gccPostfix', '')

        if gcc_bin:
            if self._ospath.exists(gcc_bin):
                self._have_tool = True
        else:
            gcc = pref + 'gcc' + postf
            gcc = self._pathutil.which(gcc)

            if gcc:
                env['gccBin'] = gcc
                self._have_tool = True




from ..runenvtool import RunEnvTool


class sshTool(RunEnvTool):
    """Secure Shell client.

Home: https://www.openssh.com/

Note:
    * sshStrictHostKeyChecking is set to "yes" by default.
      It is used by tools depending on ssh.
"""
    __slots__ = ()

    def envNames(self):
        return ['sshBin', 'sshStrictHostKeyChecking']

    def _installTool(self, env):
        self._install.deb(['openssh-client'])
        self._install.rpm(['openssh'])
        self._install.emerge(['virtual/ssh'])
        self._install.pacman(['openssh'])
        self._install.apk('openssh-client')

    def initEnv(self, env):
        env.setdefault('sshStrictHostKeyChecking', 'yes')
        super(sshTool, self).initEnv(env)




from __future__ import print_function

from ..coloring import Coloring
from ..mixins.ondemand import ext as _ext


def info(msg, label=None):
    if label:  # for backward compatibility
        infoLabel(label, msg)
        return

    line = Coloring.info('INFO: ' + msg)
    print(line, file=_ext.sys.stderr)


def infoLabel(label, msg):
    line = Coloring.infoLabel(label) + Coloring.info(msg)
    print(line, file=_ext.sys.stderr)


def warn(msg):
    print(Coloring.warn('WARNING: ' + msg), file=_ext.sys.stderr)


def errorExit(msg):
    raise RuntimeError(msg)




from ..runtimetool import RuntimeTool


class exeTool(RuntimeTool):
    """Dummy tool to execute files directly"""
    __slots__ = ()

    def initEnv(self, env):
        self._have_tool = True

    def onRun(self, config, svc, args):
        env = config['env']
        self._executil.callInteractive([svc['path']] + args)




from ..runenvtool import RunEnvTool


class unzipTool(RunEnvTool):
    """list, test and extract compressed files in a ZIP archive.
"""
    __slots__ = ()

    def _installTool(self, env):
        self._install.debrpm(['unzip'])
        self._install.emerge(['app-arch/unzip'])
        self._install.pacman(['unzip'])
        self._install.apk(['unzip'])

    def initEnv(self, env, bin_name=None):
        # Busybox's version is not enough for SDKMan
        if self._detect.isAlpineLinux() and self._ospath.islink('/usr/bin/unzip'):
            return

        super(unzipTool, self).initEnv(env, bin_name)




from ..runtimetool import RuntimeTool
from .gemtoolmixin import GemToolMixIn


class pumaTool(GemToolMixIn, RuntimeTool):
    """A ruby web server built for concurrency http://puma.io
"""
    __slots__ = ()

    def _sigReload(self):
        return self._ext.signal.SIGUSR2

    def tuneDefaults(self):
        return {
            'minMemory': '128M',
            'connMemory': '8M',
            'connFD': 8,
            'socketTypes': ['unix', 'tcp'],
            'socketType': 'unix',
            'socketProtocol': 'http',
            'scalable': True,
            'reloadable': True,
            'multiCore': False,  # use workers on service level
            'maxRequestSize': '1M',
        }

    def onRun(self, config, svc, args):
        svc_tune = svc['tune']

        #---

        if svc_tune['socketType'] == 'unix':
            socket = 'unix://{0}'.format(svc_tune['socketPath'])
        else:
            socket = 'tcp://{0}:{1}'.format(
                svc_tune['socketAddress'], svc_tune['socketPort'])

        if config['env']['type'] == 'dev':
            ruby_env = 'development'
        else:
            ruby_env = 'production'

        #---
        resource = self._ext.resource
        heap_limit = self._configutil.parseMemory(svc_tune['maxMemory'])
        # both limit RAM and HEAP (not the same)
        resource.setrlimit(resource.RLIMIT_RSS, (heap_limit, heap_limit))
        resource.setrlimit(resource.RLIMIT_DATA, (heap_limit, heap_limit))

        #---
        threads = svc_tune['maxConnections']

        puma_args = [
            '-b', socket,
            '-e', ruby_env,
            '-t', '{0}:{0}'.format(threads)
        ]

        #---

        cmd = [
            config['env']['pumaBin'],
            svc['path']
        ] + puma_args + args

        self._executil.callInteractive(cmd)




class BashToolMixIn(object):
    __slots__ = ()

    def getDeps(self):
        return ['bash']

    def _callBash(self, env, cmd=None, bash_args=[], *args, **nargs):
        if cmd:
            cmd_args = ['-c', cmd]
        else:
            cmd_args = ['-s']

        return self._executil.callExternal(
            [env['bashBin'], '--noprofile', '--norc'] + cmd_args + bash_args,
            *args, **nargs)

    def _callBashInteractive(self, env, cmd, replace=True):
        return self._executil.callInteractive([
            env['bashBin'],
            '--noprofile', '--norc',
            '-c', cmd,
        ], replace=replace)




from .ondemand import OnDemandMixIn


class DataSlots(OnDemandMixIn):
    __slots__ = (
        # ConfigMixIn
        '_startup_env',
        '_env',
        '_config',
        '_overrides',
        '_global_config',
        '_user_config',
        '_deploy_config',
        '_project_config',

        # DeployMixIn
        '_current_dir',
        '_devserve_mode',

        # LockMixIn
        '_deploy_lock',
        '_master_lock',
        '_global_lock',

        # ToolMixIn
        '_tool_impl',

        '_lastPackages',

        # ServiceMixIn
        '_running',
        '_reload_services',
        '_interruptable',
    )

    _FUTOIN_JSON = 'futoin.json'




from ..subtool import SubTool


class GemToolMixIn(SubTool):
    __slots__ = ()

    def getDeps(self):
        return ['gem', 'ruby']

    def _gemName(self):
        return self._name

    def _installTool(self, env):
        self._builddep.require(env, 'ruby')

        ver = env.get(self._name + 'Ver', None)
        version_arg = []

        if ver:
            version_arg = ['--version', ver]

        self._executil.callExternal([env['gemBin'], 'install', self._gemName(
        )] + env['gemInstallArgs'].split(' ') + version_arg)

    def updateTool(self, env):
        if env.get(self._name + 'Ver', None):
            self._installTool(env)
        else:
            self._executil.callExternal(
                [env['gemBin'], 'update', self._gemName()] + env['gemInstallArgs'].split(' '))

    def uninstallTool(self, env):
        self._executil.callExternal([env['gemBin'], 'uninstall',
                                     '--force', self._gemName()])
        self._have_tool = False




from ..runtimetool import RuntimeTool
from .sdkmantoolmixin import SdkmanToolMixIn


class scalaTool(SdkmanToolMixIn, RuntimeTool):
    """The Scala Programming Language.

Home: https://www.scala-lang.org/

Installed via SDKMan!

Requires Java >= 8.
"""
    __slots__ = ()

    def _minJava(self):
        return '8'

    def tuneDefaults(self):
        return {
            'minMemory': '256M',
            'debugOverhead': '128M',
            'connMemory': '100K',
            'debugConnOverhead': '1M',
            'socketType': 'tcp',
            'scalable': False,
        }




from .subtool import SubTool

__all__ = ['VcsTool']


class VcsTool(SubTool):
    __slots__ = ()

    def autoDetect(self, config):
        return self._autoDetectVCS(config, self.autoDetectFiles())

    def vcsGetRepo(self, config, wc_dir=None):
        raise NotImplementedError(self._name)

    def vcsCheckout(self, config, branch):
        raise NotImplementedError(self._name)

    def vcsCommit(self, config, message, files):
        raise NotImplementedError(self._name)

    def vcsTag(self, config, tag, message):
        raise NotImplementedError(self._name)

    def vcsPush(self, config, refs):
        raise NotImplementedError(self._name)

    def vcsGetRevision(self, config):
        raise NotImplementedError(self._name)

    def vcsGetRefRevision(self, config, vcs_cache_dir, branch):
        raise NotImplementedError(self._name)

    def vcsListTags(self, config, vcs_cache_dir, tag_hint):
        raise NotImplementedError(self._name)

    def vcsListBranches(self, config, vcs_cache_dir, branch_hint):
        raise NotImplementedError(self._name)

    def vcsExport(self, config, vcs_cache_dir, vcs_ref, dst_path):
        raise NotImplementedError(self._name)

    def vcsBranch(self, config, vcs_ref):
        raise NotImplementedError(self._name)

    def vcsMerge(self, config, vcs_ref, cleanup):
        raise NotImplementedError(self._name)

    def vcsDelete(self, config, vcs_cache_dir, vcs_ref):
        raise NotImplementedError(self._name)

    def vcsRevert(self, config):
        raise NotImplementedError(self._name)

    def vcsIsMerged(self, config, vcs_ref):
        raise NotImplementedError(self._name)

    def vcsClean(self, config):
        raise NotImplementedError(self._name)

    def _autoDetectVCS(self, config, vcsDir):
        if config.get('vcs', None) == self._name:
            return True

        if vcsDir in config['projectRootSet']:
            if config.get('vcs', self._name) != self._name:
                self._errorExit(
                    'Another VCS type {0} has been already detected!'.format(config['vcs']))
            config['vcs'] = self._name
            return True

        return False




from ..buildtool import BuildTool


class pipTool(BuildTool):
    """The PyPA recommended tool for installing Python packages.

Home: https://pypi.python.org/pypi/pip
"""
    __slots__ = ()

    REQUIREMENTS_FILE = 'requirements.txt'

    def autoDetectFiles(self):
        return self.REQUIREMENTS_FILE

    def getDeps(self):
        return ['python', 'virtualenv']

    def envNames(self):
        return ['pipBin', 'pipVer']

    def _installTool(self, env):
        ospath = self._ospath

        if ospath.exists(env['pipBin']):
            self.updateTool(env)
        else:
            self._executil.callExternal([
                ospath.join(env['virtualenvDir'], 'bin',
                            'easy_install'), 'pip'
            ])

    def updateTool(self, env):
        self._executil.callExternal([
            env['pipBin'], 'install', '-q',
            '--upgrade',
            'pip>={0}'.format(env['pipVer']),
        ])

    def uninstallTool(self, env):
        pass

    def initEnv(self, env):
        ospath = self._ospath
        pipBin = ospath.join(env['virtualenvDir'], 'bin', 'pip')
        pipBin = env.setdefault('pipBin', pipBin)
        pipVer = env.setdefault('pipVer', '9.0.1')

        if ospath.exists(pipBin):
            pipFactVer = self._executil.callExternal(
                [pipBin, '--version'], verbose=False)
            pipFactVer = [int(v) for v in pipFactVer.split(' ')[1].split('.')]
            pipNeedVer = [int(v) for v in pipVer.split('.')]

            self._have_tool = pipNeedVer <= pipFactVer

    def onPrepare(self, config):
        if self._ospath.exists(self.REQUIREMENTS_FILE):
            self._executil.callExternal(
                [config['env']['pipBin'], 'install', '-r', self.REQUIREMENTS_FILE])




from ..subtool import SubTool


class NpmToolMixIn(SubTool):
    __slots__ = ()

    def getDeps(self):
        return ['node', 'npm']

    def _npmName(self):
        return self._name

    def _isGlobalNpm(self):
        return self._detect.isAlpineLinux()

    def _installTool(self, env):
        cmd = [env['npmBin'], 'install', '-g', self._npmName()]

        if self._isGlobalNpm():
            self._executil.trySudoCall(cmd)
        else:
            self._executil.callExternal(cmd)

    def updateTool(self, env):
        cmd = [env['npmBin'], 'update', '-g', self._npmName()]

        if self._isGlobalNpm():
            self._executil.trySudoCall(cmd)
        else:
            self._executil.callExternal(cmd)

    def uninstallTool(self, env):
        cmd = [env['npmBin'], 'uninstall', '-g', self._npmName()]

        if self._isGlobalNpm():
            self._executil.trySudoCall(cmd)
        else:
            self._executil.callExternal(cmd)

        self._have_tool = False




from ..runenvtool import RunEnvTool


class gzipTool(RunEnvTool):
    """Compression utility designed to be a replacement for compress.

Home: http://www.gzip.org/
"""
    __slots__ = ()

    def _installTool(self, env):
        self._install.debrpm(['gzip'])
        self._install.emerge(['app-arch/gzip'])
        self._install.pacman(['gzip'])
        self._install.apk(['gzip'])
        self._install.brew('gzip')

    def initEnv(self, env, bin_name=None):
        # Busybox's version is not enough for SDKMan
        if self._detect.isAlpineLinux() and self._ospath.islink('/bin/gzip'):
            return

        super(gzipTool, self).initEnv(env, bin_name)




from ..runenvtool import RunEnvTool


class tarTool(RunEnvTool):
    """The GNU version of the tar archiving utility.
"""
    __slots__ = ()

    def _installTool(self, env):
        self._install.debrpm(['tar'])
        self._install.emerge(['app-arch/tar'])
        self._install.pacman(['tar'])
        self._install.apk(['tar'])
        self._install.brew('gnu-tar')

    def initEnv(self, env, bin_name=None):
        # Busybox's version is not enough for SDKMan
        if self._detect.isAlpineLinux() and self._ospath.islink('/bin/tar'):
            return

        super(tarTool, self).initEnv(env, bin_name)




from ..buildtool import BuildTool
from ..rmstool import RmsTool


class npmTool(BuildTool, RmsTool):
    """npm is the package manager for JavaScript.

Home: https://www.npmjs.com/

In RMS mode support tuning through .toolTune.npm:
    .tag - npm publish --tag
    .access - npm publish --access

Note: it auto-disables, if Yarn tool is detected
"""
    __slots__ = ()

    PACKAGE_JSON = 'package.json'

    def autoDetectFiles(self):
        return self.PACKAGE_JSON

    def getDeps(self):
        return ['node']

    def _isGlobalNpm(self):
        return self._detect.isAlpineLinux()

    def _isYarnInUse(self, config):
        return 'yarn' in config['toolOrder']

    def _installTool(self, env):
        if self._isGlobalNpm():
            if env['nodeVer'] == 'current':
                self._install.apk('nodejs-npm-current')
            else:
                self._install.apk('nodejs-npm')
            return

    def updateTool(self, env):
        if self._isGlobalNpm():
            return

        self._executil.callExternal([env['npmBin'], 'update', '-g', 'npm'])

    def uninstallTool(self, env):
        pass

    def loadConfig(self, config):
        content = self._pathutil.loadJSONConfig(self.PACKAGE_JSON)
        if content is None:
            return

        for f in ('name', 'version'):
            if f in content and f not in config:
                config[f] = content[f]

    def updateProjectConfig(self, config, updates):
        def updater(json):
            for f in ('name', 'version'):
                if f in updates:
                    json[f] = updates[f]

        return self._pathutil.updateJSONConfig(self.PACKAGE_JSON, updater)

    def onPrepare(self, config):
        if self._ospath.exists(self.PACKAGE_JSON) and not self._isYarnInUse(config):
            npmBin = config['env']['npmBin']
            self._executil.callExternal([npmBin, 'install'])

    def onPackage(self, config):
        npmBin = config['env']['npmBin']

        if self._ospath.exists(self.PACKAGE_JSON) and not self._isYarnInUse(config):
            self._executil.callExternal([npmBin, 'prune', '--production'])

        if config.get('rms', None) == self._name:
            self._executil.callExternal([npmBin, 'pack'])
            package = '{0}-{1}.tgz'.format(
                config['name'], config['version'])
            self._pathutil.addPackageFiles(config, package)

    def rmsUpload(self, config, rms_pool, package_list):
        npmBin = config['env']['npmBin']
        cmd = [npmBin, 'publish']

        tune = config.get('toolTune', {}).get('npm', {})

        if 'tag' in tune:
            cmd += ['--tag', tune['tag']]

        if 'access' in tune:
            cmd += ['--access', tune['access']]

        self._executil.callExternal(cmd)

    def rmsPoolList(self, config):
        return [
            'registry',
        ]




from ...mixins.ondemand import ext as _ext

_pacman = _ext.pathutil.which('pacman')


def pacman(packages):
    if _pacman:
        packages = _ext.configutil.listify(packages)

        _ext.executil.trySudoCall(
            [_pacman, '-S', '--noconfirm', '--needed'] + packages,
            errmsg='you may need to install the build deps manually !'
        )




#==================


def timeouts(env, fmt):
    timeouts = env['timeouts']
    connect_to = int(timeouts['connect'])
    read_to = int(timeouts['read'])
    total_to = int(timeouts['total'])

    if fmt == 'requests':
        return (connect_to, read_to)
    elif fmt == 'curl':
        return [
            '--connect-timeout', '{0}'.format(int(connect_to)),
            '--max-time', '{0}'.format(int(total_to)),
        ]

    raise NotImplementedError('Unknown format: {0}'.format(fmt))


#==================
_memory_mult_table = {
    'B': 1,
    'K': (1 << 10),
    'M': (1 << 20),
    'G': (1 << 30),
}


def parseMemory(val):
    b = int(val[:-1])
    m = _memory_mult_table[val[-1]]

    if b <= 0:
        raise ValueError('Memory must be positive')

    return b * m


def toMemory(val):
    res = None
    old_v = 0

    for (k, v) in _memory_mult_table.items():
        if val % v:
            continue
        elif v > old_v:
            res = '{0}{1}'.format(int(val / v), k)
            old_v = v

    return res
#==================


def listify(val):
    if isinstance(val, list):
        return val

    return [val]




from ..buildtool import BuildTool
from ..testtool import TestTool
from .sdkmantoolmixin import SdkmanToolMixIn


class sbtTool(SdkmanToolMixIn, BuildTool, TestTool):
    """The interactive build tool (Scala).

Home: http://www.scala-sbt.org/

Installed via SDKMan!

First run of SBT may consume a lot of time on post-installation steps.

Build targets:
    prepare -> clean
    build -> compile
    package -> package
    check -> test
Override targets with .config.toolTune.

Requires Java >= 8.
"""
    __slots__ = ()

    def _minJava(self):
        return '8'

    def autoDetectFiles(self):
        return 'build.sbt'

    def onPrepare(self, config):
        target = self._getTune(config, 'prepare', 'clean')
        self._executil.callExternal([config['env']['sbtBin'], target])

    def onBuild(self, config):
        target = self._getTune(config, 'build', 'compile')
        self._executil.callExternal([config['env']['sbtBin'], target])

    def onPackage(self, config):
        target = self._getTune(config, 'package', 'package')
        self._executil.callExternal([config['env']['sbtBin'], target])
        self._pathutil.addPackageFiles(config, 'target/scala-*/*.jar')

    def onRunDev(self, config):
        target = self._getTune(config, 'run', 'check')
        self._executil.callExternal([config['env']['sbtBin'], target])

    def onCheck(self, config):
        target = self._getTune(config, 'check', 'test')
        self._executil.callExternal([config['env']['sbtBin'], target])




from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn


class gulpTool(NpmToolMixIn, BuildTool):
    """Automate and enhance your workflow (Node.js).

Home: http://gulpjs.com/    
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return 'gulpfile.js'

    def onBuild(self, config):
        self._executil.callExternal([config['env']['gulpBin']])




from ..buildtool import BuildTool
from .piptoolmixin import PipToolMixIn


class cidTool(PipToolMixIn, BuildTool):
    "Noop FutoIn-CID - a workaround to allow CID use from virtualenv"

    __slots__ = ()

    def _pipName(self):
        return 'futoin-cid'

    def _installTool(self, env):
        source_dir = self._environ.get('CID_SOURCE_DIR', None)

        if source_dir:
            self._executil.callExternal(
                [env['pipBin'], 'install', '-q', '-e', source_dir])
        else:
            PipToolMixIn._installTool(self, env)




from ..buildtool import BuildTool


class cmakeTool(BuildTool):
    """Build, Test and Package Your Software With CMake.

Home: https://cmake.org/

CMake creates a build folder and does all processing in it.
Build folder is configurable through cmakeBuildDir env.
"""
    __slots__ = ()

    def getOrder(self):
        return -10

    def autoDetectFiles(self):
        return 'CMakeLists.txt'

    def envNames(self):
        return ['cmakeBin', 'cmakeBuildDir']

    def _installTool(self, env):
        self._install.deb(['build-essential'])
        self._install.rpm(['gcc', 'gcc-c++'])
        self._install.pacman(['gcc'])
        self._builddep.essential()

        self._install.debrpm(['cmake'])
        self._install.emerge(['dev-util/cmake'])
        self._install.pacman(['cmake'])
        self._install.apk(['cmake'])
        self._install.brew('cmake')

    def initEnv(self, env):
        env.setdefault('cmakeBuildDir', 'build')
        super(cmakeTool, self).initEnv(env)

    def onPrepare(self, config):
        build_dir = config['env']['cmakeBuildDir']
        self._pathutil.rmTree(build_dir)

    def onBuild(self, config):
        ospath = self._ospath
        os = self._os
        build_dir = config['env']['cmakeBuildDir']

        if ospath.exists(build_dir):
            self._executil.callExternal([config['env']['cmakeBin'], build_dir])
        else:
            os.mkdir(build_dir)
            os.chdir(build_dir)
            self._executil.callExternal(
                [config['env']['cmakeBin'], config['wcDir']])
            os.chdir(config['wcDir'])

        self._executil.callExternal(
            [config['env']['cmakeBin'], '--build', build_dir])




from .subtool import SubTool

__all__ = ['BuildTool']


class BuildTool(SubTool):
    __slots__ = ()

    def onPrepare(self, config):
        pass

    def onBuild(self, config):
        pass

    def onPackage(self, config):
        pass




from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn


class gruntTool(NpmToolMixIn, BuildTool):
    """Grunt: The JavaScript Task Runner.

Home: https://gruntjs.com/    
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return ['Gruntfile.js', 'Gruntfile.coffee']

    def _npmName(self):
        return 'grunt-cli'

    def onBuild(self, config):
        gruntBin = config['env']['gruntBin']
        self._executil.callExternal([gruntBin])




from ..runenvtool import RunEnvTool


class xzTool(RunEnvTool):
    """Free general-purpose data compression software with a high compression ratio.

Home: https://tukaani.org/xz/
"""
    __slots__ = ()

    def _installTool(self, env):
        self._install.debrpm(['xz-utils'])
        self._install.emerge(['app-arch/xz-utils'])
        self._install.pacman(['xz-utils'])
        self._install.apk(['xz'])
        self._install.brew('xz')




from ..mixins.ondemand import ext as _ext


def sort(verioned_list):
    def castver(v):
        res = _ext.re.split(r'[\W_]+', v)
        for (i, vc) in enumerate(res):
            try:
                res[i] = int(vc, 10)
            except:
                pass
        return res

    verioned_list.sort(key=castver)


def latest(verioned_list):
    sort(verioned_list)
    return verioned_list[-1]




from ..buildtool import BuildTool
from ..runenvtool import RunEnvTool
from .piptoolmixin import PipToolMixIn


class dockercomposeTool(PipToolMixIn, BuildTool, RunEnvTool):
    """Compose is a tool for defining and running multi-container Docker applications.

Home: https://docs.docker.com/compose/

Experimental support.
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return ['docker-compose.yml', 'docker-compose.yaml']

    def getDeps(self):
        return ['pip', 'docker']

    def getOrder(self):
        return 20

    def _pipName(self):
        return 'docker-compose'

    def _installTool(self, env):
        self._requirePythonDev(env)
        super(dockercomposeTool, self)._installTool(env)

    def onBuild(self, config):
        self._executil.callExternal(
            [config['env']['dockercomposeBin'], 'build'])

    def onPackage(self, config):
        self._executil.callExternal(
            [config['env']['dockercomposeBin'], 'bundle'])

    def initEnv(self, env):
        super(dockercomposeTool, self).initEnv(env, 'docker-compose')




from .subtool import SubTool

__all__ = ['RuntimeTool']


class RuntimeTool(SubTool):
    __slots__ = ()
    DEFAULT_EXIT_TIMEOUT = 5000

    def __init__(self, name):
        super(RuntimeTool, self).__init__(name)

    def onRun(self, config, svc, args):
        env = config['env']
        self._executil.callInteractive([
            env[self._name + 'Bin'], svc['path']
        ] + args)

    def tuneDefaults(self):
        return {
            'minMemory': '1M',
            'connMemory': '1M',
            'scalable': False,
            'reloadable': False,
            'multiCore': False,
            'exitTimeoutMS': self.DEFAULT_EXIT_TIMEOUT,
            'maxRequestSize': '1M',
            'socketProtocol': 'custom',
        }

    def onStop(self, config, pid, tune):
        self._signalPID(pid, self._sigStop())

    def onReload(self, config, pid, tune):
        if tune['reloadable']:
            self._signalPID(pid, self._sigReload())
        else:
            self.onStop(config, pid, tune)

    def onPreConfigure(self, config, runtime_dir, svc, cfg_svc_tune):
        pass

    def _signalPID(self, pid, sig):
        self._os.kill(pid, sig)

    def _sigReload(self):
        return self._ext.signal.SIGHUP

    def _sigStop(self):
        return self._ext.signal.SIGTERM




from functools import wraps as _wraps


def simple_memo(f):
    @_wraps(f)
    def wrap():
        fdict = f.__dict__

        if 'cached_val' in fdict:
            return fdict['cached_val']

        val = f()
        fdict['cached_val'] = val
        return val

    return wrap


def complex_memo(f):
    f.cached = {}

    @_wraps(f)
    def wrap(*args, **kwargs):
        cached = f.cached
        key = str(args) + str(kwargs)
        if key in cached:
            return cached[key]

        val = f(*args, **kwargs)

        if val is not None:
            cached[key] = val

        return val

    return wrap
